/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WF0EL_H
#define WF0EL_H

#include "flow.h"
#include "wObserver.h"
#include "wElement.h"
#include <vector>

namespace flow
{

/**
 * @brief Scalar function to be integrated over an element
 */
class FLOW_API F0El : public fwk::Observer
{
public:
    F0El() : Observer() {}
    virtual ~F0El() {}

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const = 0;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const = 0;
};

/**
 * @brief Constant function
 */
class FLOW_API F0ElC : public F0El
{
    double v;

public:
    F0ElC(double _v) : F0El(), v(_v) {}
    virtual void update(double _v) override;

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

/**
 * @brief Nonlinear density
 *
 * The density is limited (Padé approximation) for large value of velocity and
 * particularized for diatomic gas. The input _mC2 is the square of the critical
 * Mach number used to compute the critical velocity.
 */
class FLOW_API F0ElRho : public F0El
{
    double gamma;  ///< heat capacity ratio (diatomic gas only)
    double mInf;   ///< freestream Mach number
    double gradUC; ///< critical velocity

public:
    F0ElRho(double _mInf, double _mC2 = 5) : F0El(), gamma(1.4), mInf(_mInf)
    {
        gradUC = sqrt(_mC2 * (1 / (mInf * mInf) + (gamma - 1) / 2) / (1 + (gamma - 1) / 2 * _mC2));
    }

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

/**
 * @brief Linear density (constant)
 */
class FLOW_API F0ElRhoL : public F0El
{
public:
    F0ElRhoL() : F0El() {}

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

/**
 * @brief Nonlinear Mach number
 *
 * Particularized for diatomic gas
 */
class FLOW_API F0ElMach : public F0El
{
    double gamma; ///< heat capacity ratio (diatomic gas only)
    double mInf;  ///< freestream Mach number

public:
    F0ElMach(double _mInf) : F0El(), gamma(1.4), mInf(_mInf) {}

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

/**
 * @brief Linear Mach number (constant)
 */
class FLOW_API F0ElMachL : public F0El
{
public:
    F0ElMachL() : F0El() {}

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

/**
 * @brief Nonlinear pressure coefficient
 *
 * Particularized for diatomic gas
 */
class FLOW_API F0ElCp : public F0El
{
    double gamma; ///< heat capacity ratio (diatomic gas only)
    double mInf;  ///< freestream Mach number

public:
    F0ElCp(double _mInf) : F0El(), gamma(1.4), mInf(_mInf) {}

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

/**
 * @brief Linear pressure coefficient
 */
class FLOW_API F0ElCpL : public F0El
{

public:
    F0ElCpL() : F0El() {}

    virtual double eval(tbox::Element const &e, std::vector<double> const &u,
                        size_t k) const override;
    virtual double evalGrad(tbox::Element const &e, std::vector<double> const &u,
                            size_t k) const override;
};

} // namespace flow

#endif //WF0EL_H
