/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wPotentialResidual.h"
#include "wMedium.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"

using namespace tbox;
using namespace flow;

/**
 * @brief Build the residual matrix for a fixed-point iteration, on one element
 *
 * A = \int rho * grad_phi * grad_psi dV
 */
Eigen::MatrixXd PotentialResidual::buildFixed(Element const &e, std::vector<double> const &phi, Medium const &fluid)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Build
    Eigen::MatrixXd A = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions gradient
        Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
        // Elementary stiffness matrix
        A += dSf.transpose() * dSf * fluid.rho->eval(e, phi, k) * gauss.getW(k) * e.getDetJ(k);
    }
    return A;
}

/**
 * @brief Build the residual vector, on one element
 *
 * b = \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV
 */
Eigen::VectorXd PotentialResidual::build(Element const &e, Element const &eU, std::vector<double> const &phi, Medium const &fluid, double muC, double mCO)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Subsonic contribution
    Eigen::VectorXd b = Eigen::VectorXd::Zero(e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
        b += fluid.rho->eval(e, phi, k) * (e.getJinv(k) * cache.getDsf(k)).transpose() * e.computeGradient(phi, k) * gauss.getW(k) * e.getDetJ(k);

    // Supersonic contribution
    double mach = fluid.mach->eval(e, phi, 0);
    if (mach > mCO)
    {
        double mu = muC * (1 - (mCO * mCO) / (mach * mach)); // switching  function
        double rhoU = fluid.rho->eval(eU, phi, 0);           // density on upwind element
        // scale 1st term
        b *= 1 - mu;
        // contribution of upwind element
        for (size_t k = 0; k < gauss.getN(); ++k)
            b += mu * rhoU * (e.getJinv(k) * cache.getDsf(k)).transpose() * e.computeGradient(phi, k) * gauss.getW(k) * e.getDetJ(k);
    }
    return b;
}

/**
 * @brief Build the gradient of the residual vector with respect to the flow variable (jacobian), on one element
 *
 * A = \int d( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV
 *   = \int (1-mu)*drho * grad_phi * grad_psi dV
 *   + \int mu*drhoU * grad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * dgrad_phi * grad_psi dV
 *   + \int (rhoU - rho)*dmu * grad_phi * grad_psi dV
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> PotentialResidual::buildGradientFlow(tbox::Element const &e, tbox::Element const &eU, std::vector<double> const &phi, Medium const &fluid, double muC, double mCO)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Subsonic contribution
    Eigen::MatrixXd A1 = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Gauss point and determinant
        double wdj = gauss.getW(k) * e.getDetJ(k);
        // Shape functions and solution gradient
        Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
        Eigen::VectorXd dPhi = e.computeGradient(phi, k);

        // rho * grad_phi*grad_psi + drho * grad_phi*grad_psi
        A1 += fluid.rho->eval(e, phi, k) * dSf.transpose() * dSf * wdj;
        A1 += fluid.rho->evalGrad(e, phi, k) * dSf.transpose() * dPhi * dPhi.transpose() * dSf * wdj;
    }

    // Supersonic contribution
    Eigen::MatrixXd A2;
    double mach = fluid.mach->eval(e, phi, 0);
    if (mach > mCO)
    {
        // switching function and gradient
        double mu = muC * (1 - (mCO * mCO) / (mach * mach));
        double dmu = 2 * muC * mCO * mCO / (mach * mach * mach);
        // density and gradients on upwind elements
        Eigen::VectorXd dPhiU = eU.computeGradient(phi, 0);
        double rhoU = fluid.rho->eval(eU, phi, 0);
        double dRhoU = fluid.rho->evalGrad(eU, phi, 0);
        Eigen::MatrixXd const &dSfU = eU.getJinv(0) * eU.getVCache().getDsf(0);

        // upwinding
        A1 *= 1 - mu;
        A2 = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Gauss point and determinant
            double wdj = gauss.getW(k) * e.getDetJ(k);
            // Shape functions and solution gradient
            Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
            Eigen::VectorXd dPhi = e.computeGradient(phi, k);

            // mu*drhoU*grad_phi*grad_psi
            A2 += mu * dRhoU * dSf.transpose() * dPhi * dPhiU.transpose() * dSfU * wdj;
            // mu*rhoU*grad_phi*grad_psi + dmu*(rhoU-rho)*grad_phi*grad_psi
            A1 += mu * rhoU * dSf.transpose() * dSf * wdj;
            A1 += dmu * (rhoU - fluid.rho->eval(e, phi, k)) * fluid.mach->evalGrad(e, phi, k) * dSf.transpose() * dPhi * dPhi.transpose() * dSf * wdj;
        }
    }
    return std::make_tuple(A1, A2);
}

/**
 * @brief Build the gradient of the residual vector with respect to the nodes, on one element
 *
 * A = d( \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV )
 *   = \int (1-mu)*drho * grad_phi * grad_psi dV
 *   + \int mu*drhoU * grad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * dgrad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * dgrad_psi dV
 *   + \int (rhoU - rho)*dmu * grad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi ddV
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> PotentialResidual::buildGradientMesh(tbox::Element const &e, tbox::Element const &eU, std::vector<double> const &phi, Medium const &fluid, int nDim, double muC, double mCO)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Subsonic contribution
    Eigen::MatrixXd A1 = Eigen::MatrixXd::Zero(e.nodes.size(), nDim * e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Density and gradients
        Eigen::VectorXd dPhi = e.computeGradient(phi, k);
        double rho = fluid.rho->eval(e, phi, k);
        double dRho = fluid.rho->evalGrad(e, phi, k);
        Eigen::MatrixXd const &dSf = cache.getDsf(k);
        Eigen::MatrixXd const &iJ = e.getJinv(k);
        // Jacobian gradients
        std::vector<Eigen::MatrixXd> const &dJ = e.getGradJ(k);
        std::vector<double> const &dDetJ = e.getGradDetJ(k);
        // Gauss points and determinant
        double w = gauss.getW(k);
        double detJ = e.getDetJ(k);
        for (size_t i = 0; i < e.nodes.size(); ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                A1.col(idx) += w * dRho * dPhi.transpose() * (-iJ * dJ[idx] * dPhi) * (iJ * dSf).transpose() * dPhi * detJ; // drho * grad_phi*grad_psi * detj
                A1.col(idx) += w * rho * (iJ * dSf).transpose() * (-iJ * dJ[idx] * dPhi) * detJ;                            // rho * dgrad_phi * grad_psi * detj
                A1.col(idx) += w * rho * (-iJ * dJ[idx] * iJ * dSf).transpose() * dPhi * detJ;                              // rho * grad_phi * dgrad_psi * detj
                A1.col(idx) += w * rho * (iJ * dSf).transpose() * dPhi * dDetJ[idx];                                        // rho * grad_phi * grad_psi * ddetj
            }
        }
    }
    // Supersonic contribution
    Eigen::MatrixXd A2;
    double mach = fluid.mach->eval(e, phi, 0);
    if (mach > mCO)
    {
        // switching function and gradient
        double mu = muC * (1 - (mCO * mCO) / (mach * mach));
        double dmu = 2 * muC * mCO * mCO / (mach * mach * mach);
        // upwind density and gradient
        Eigen::VectorXd dPhiU = eU.computeGradient(phi, 0);
        double rhoU = fluid.rho->eval(eU, phi, 0);
        double dRhoU = fluid.rho->evalGrad(eU, phi, 0);
        Eigen::MatrixXd const &iJU = eU.getJinv(0);
        std::vector<Eigen::MatrixXd> const &dJU = eU.getGradJ(0);

        // upwinding
        A1 *= 1 - mu;
        A2 = Eigen::MatrixXd::Zero(e.nodes.size(), nDim * eU.nodes.size());
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Density, Mach and Gradients
            Eigen::VectorXd dPhi = e.computeGradient(phi, k);
            double rho = fluid.rho->eval(e, phi, k);
            double dM = fluid.mach->evalGrad(e, phi, k);
            Eigen::MatrixXd const &dSf = cache.getDsf(k);
            Eigen::MatrixXd const &iJ = e.getJinv(k);
            // Jacobian gradients
            std::vector<Eigen::MatrixXd> const &dJ = e.getGradJ(k);
            std::vector<double> const &dDetJ = e.getGradDetJ(k);
            // Gauss points and determinant
            double w = gauss.getW(k);
            double detJ = e.getDetJ(k);
            for (size_t m = 0; m < nDim; ++m)
            {
                for (size_t i = 0; i < e.nodes.size(); ++i)
                {
                    size_t idx = i * nDim + m;
                    A1.col(idx) += w * mu * rhoU * (iJ * dSf).transpose() * (-iJ * dJ[idx] * dPhi) * detJ;                                          // mu * rhoU * dgrad_phi * grad_psi * detj
                    A1.col(idx) += w * mu * rhoU * (-iJ * dJ[idx] * iJ * dSf).transpose() * dPhi * detJ;                                            // mu * rhoU * grad_phi * dgrad_psi * detj
                    A1.col(idx) += w * mu * rhoU * (iJ * dSf).transpose() * dPhi * dDetJ[idx];                                                      // mu * rhoU * grad_phi * grad_psi * ddetj
                    A1.col(idx) += w * (-rho + rhoU) * dmu * dM * dPhi.transpose() * (-iJ * dJ[idx] * dPhi) * (iJ * dSf).transpose() * dPhi * detJ; // (rhoU-rho) * dmu * grad_phi * grad_psi * detj
                }
                for (size_t i = 0; i < eU.nodes.size(); ++i)
                {
                    size_t idx = i * nDim + m;
                    A2.col(idx) += w * mu * dRhoU * dPhiU.transpose() * (-iJU * dJU[idx] * dPhiU) * (iJ * dSf).transpose() * dPhi * detJ; // mu * drhoU * grad_phi * grad_psi * detj
                }
            }
        }
    }
    return std::make_tuple(A1, A2);
}
