#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute flow around the Onera M6 wing at 3 degrees AOA and Mach 0.84
# Adrien Crovato

import numpy as np
import flow.utils as floU
import flow.default as floD
import tbox.utils as tbxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

try:
   import tboxVtk
   BestWriter = tboxVtk.VtkExport
   canPost = True
except:
   BestWriter = tbox.GmshExport
   canPost = False

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3.06*np.pi/180
    M_inf = 0.839
    dim = 3

    # define dimension and mesh size
    lgt = 3.0 # channel half length
    hgt = 3.0 # channel half height
    wdt = 3.0 # channel width
    S_ref = 0.7528 # reference area
    c_ref = 0.64 # reference chord (MAC)
    fms = 0.5 # farfield mesh size
    rlems = 0.004 # root leading edge mesh size
    rtems = 0.008 # root trailing edge mesh size
    tlems = 0.002 # tip leading mesh size
    ttems = 0.008 # tip trailing mesh size

    # mesh the wing
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msLeRt' : rlems, 'msTeRt' : rtems, 'msLeTp' : tlems, 'msTeTp' : ttems, 'msF' : fms}
    msh, _ = floD.mesh(dim, 'models/oneraM6.geo', pars, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    writer = BestWriter(msh)
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, _, _ = floD.problem(msh, dim, alpha, 0., M_inf, S_ref, c_ref, 0., 0., 0., 'wing', tp = 'teTip')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    solver.run()
    solver.save(writer)
    tms['solver'].stop()

    # post process
    tms['post'].start()
    if canPost:
        floU.writeSlices(msh.name,[0.01, 0.239, 0.526, 0.777, 0.957, 1.076, 1.136, 1.184],5)
        data = tbxU.read('slice_2.dat')
        tbxU.plot(data[:,3], data[:,4], 'x', 'Cp', 'Pressure distribution at MAC', True)
    tms['post'].stop()

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('CL', solver.Cl, 0.29, 1e-1))
    tests.add(CTest('CD', solver.Cd, 0.011, 1e-1))
    tests.add(CTest('CS', solver.Cs, 0.014, 1e-1))
    tests.add(CTest('CM', solver.Cm, -0.212, 1e-1))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
